from td import *  # pyright: ignore[reportMissingImports]
from typing import Dict, List, Literal

def table_to_dicts( table_op:tableDAT, key_source:Literal["row", "col"] = "row" ) -> List[Dict[str, str]]:
    keys = tuple( cell.val for cell in getattr(table_op, key_source)(0) )
    collection_source = getattr(table_op, f"{key_source}s")
    return [ {
        keys[ index ] :  cell.val for index ,cell in enumerate(collection)  
    } for collection in collection_source()[1:] ]

def append_dict_to_table( table_op:tableDAT, source_dict:dict, key_source:Literal["row", "col"] = "row", default_value:str = "" ):
    keys = tuple( cell.val for cell in getattr(table_op, key_source)(0) )
    getattr( table_op, f"append{key_source.capitalize()}")([
        source_dict.get( key, default_value) for key in keys
    ])

def pivot_table(input_dat:tableDAT, script_op:scriptDAT, key_source:Literal["row", "col"] = "col", id_source:str = "path", names_source:str = "name", value_source = "value"):
    """
        Thanks Mickey!
        Takes the input of our scriptDat and pivots it based on the source-arguments.
        The default arguments are set for multi-parameter tables.
    """
    
    script_op.clear()
    
    # Extract unique paths and parameter names

    item_ids = [cell.val for cell in getattr(input_dat, key_source)(id_source)[1:]]
    unique_paths = list(dict.fromkeys(item_ids))  # Preserves order, removes duplicates
    
    result_name = [cell.val for cell in getattr(input_dat, key_source)(names_source)[1:]]
    unique_params = list(dict.fromkeys(result_name))
    
    script_op.setSize(len(unique_paths), len(unique_params))
    
    # Add header row
    script_op.insertRow(unique_params, -1)
    script_op.insertCol([id_source], -1)
    
    for row_idx, path in enumerate(unique_paths, start=1):
        script_op[row_idx, id_source] = path # pyright: ignore[reportIndexIssue]
        
        # Find all rows in input that match this path
        matching_rows = input_dat.findCells(path, cols=[id_source])
        
        # Fill in parameter values for this path
        for match in matching_rows:
            # Derivative should add the correct __getitem__ method in the stups.

            param_name = input_dat[match.row, names_source]     # pyright: ignore[reportIndexIssue]
            param_value = input_dat[match.row, value_source]    # pyright: ignore[reportIndexIssue]
            script_op[row_idx, param_name] = param_value        # pyright: ignore[reportIndexIssue]


from td import *  # pyright: ignore[reportMissingImports]
from typing import Dict, Callable, Tuple, List, Set
from .operator import iter_parents

def _empty_callback(_source:OP, emitter:OP, event_name:str, bubbled:bool, *args, **kwargs):
    pass

EventCallbackType = Callable[[OP, OP, str, bool, Tuple[any, ...], Dict[any, any]], None]


_subscriptions:Dict[str, Dict[str, Set[EventCallbackType]]] = {} 


def subscribe( emitter:OP, event:str, callback:EventCallbackType):
    _subscriptions.setdefault( emitter.path , {}).setdefault(event,  set()).add( callback )

def unsubscribe( emitter:OP, event:str, callback:EventCallbackType):
    _subscriptions.setdefault( emitter.path , {}).setdefault(event,  set()).remove( callback )

def emit( _emitter:OP, _event:str, *args, _bubble = True, **kwargs):
    for _bubble in [_emitter] + list(iter_parents( _emitter )):
        for callback in _subscriptions.get( _bubble.path, {} ).get( _event, []):
            try:
                callback( _emitter, _bubble, _event, _emitter != _bubble, *args, **kwargs )
            except Exception as e:
                debug("Encountered exception during event emitting.", _emitter, _bubble, _event, e)



from td import * # pyright: ignore[reportMissingImports]

from .timing import Ticker, Timer

from typing import Union, List, Coroutine

import asyncio
from asyncio import Task

loop = asyncio.new_event_loop()
loop.stop()

def _async_tick( timer:Timer):
    loop.stop()
    loop.run_forever()

_async_ticker = Ticker([ _async_tick ])


def execute(coroutines:Union[ List[Coroutine], Coroutine]) -> List[Task]:
	"""
		Runs all routines concurrently and returns a list of tasks.
	"""
	returnTasks = []
	if not isinstance( coroutines, list): coroutines = [coroutines]
		
	for coroutine in coroutines:
		returnTasks.append( 
			loop.create_task(coroutine)
		)
	return returnTasks

def cancel(killList:List[Task] = [] ):
	"""
		Cancels all tasks currently active or the defines task in the list.
	"""
	for task in killList or asyncio.all_tasks(loop):
		task.cancel()


from td import * # pyright: ignore[reportMissingImports]
from typing import List

def quantize_attribute(target:object, quantize_step:int, attribute_name:str):
    target_attribute = getattr( target, attribute_name)
    if not isinstance( target_attribute, (float, int)): 
        raise TypeError(f"Expected Int or Float argument, got {type(target_attribute)}")
    new_value = round(target_attribute/quantize_step)*quantize_step
    setattr( target, attribute_name, new_value )
    return new_value

def quantize_operator_position( target_op:OP, quantize_step:int):
    quantize_attribute( target_op, quantize_step, "nodeX")
    quantize_attribute( target_op, quantize_step, "nodeY")


def quantize_operator_size( target_op:OP, quantize_step:int):
    quantize_attribute( target_op, quantize_step, "nodeWidth")
    quantize_attribute( target_op, quantize_step, "nodeHeight")

def quantize_annoations_comps(target_comp:COMP, quantize_step:int):
    for child in target_comp.findChildren(type = annotateCOMP):
        quantize_operator_position( child, quantize_step)
        quantize_operator_size( child, quantize_step)


def align_vertical( target_ops:List[OP], distance = 120, start_x = None, start_y = None ):
    start_x = target_ops[0].nodeX if start_x is None else start_x
    start_y = target_ops[0].nodeY if start_y is None else start_y

    for index, operator in enumerate(sorted(target_ops, key = lambda operator: operator.nodeX)):
        operator.nodeX = start_x
        operator.nodeY = start_y + distance * index


def align_horizontal( target_ops:List[OP], distance = 120, start_x = None, start_y = None ):
    start_x = target_ops[0].nodeX if start_x is None else start_x
    start_y = target_ops[0].nodeY if start_y is None else start_y

    for index, operator in enumerate(sorted(target_ops, key = lambda operator: operator.nodeX)):
        operator.nodeX = start_x + distance * index
        operator.nodeY = start_y 


from td import *  # pyright: ignore[reportMissingImports]

__minimum_td_version__ = "2023.1200"

from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version("touchutilcollection")
except PackageNotFoundError:
    __version__ = "DEV"

import logging
from os import environ

logger = logging.getLogger()
log_level = getattr(logging, environ.get("TOUCHLAUNCH_LOGLEVEL", "INFO"), None) or logging.INFO
logging.basicConfig(level=log_level)

if float(app.build) < float(__minimum_td_version__):
    logger.warning(f"{__minimum_td_version__} required, found {app.build}")


from td import * # pyright: ignore[reportMissingImports]

from functools import lru_cache
from typing import Union

@lru_cache( maxsize = None )
def get_parameter_page(operator:OP, pagename:str, builtin:bool = True, custom:bool = True) -> Union[Page, None]:
	for page in operator.pages * builtin + operator.customPages * custom:
		if page.name == pagename: return page

from typing import List, Dict, Any

class MenuSource:
	def __init__(self, names:List[str], labels:List[str] = []) -> None:
		if labels and len(labels) != len(names):
			raise TypeError("Names and labels need same length.")
		self.names = names
		self.labels = labels or names
	
	@property
	def menuLabels(self):
		return self.labels
	
	@property
	def menuNames(self):
		return self.names
	
def dictkeys_to_source(source_dict:Dict[str, Any], capitalize_labels=False):
    return MenuSource(
		[ key for key in source_dict.keys() ],
		[ key.capitalize() for key in source_dict.keys() ] if capitalize_labels else []
	)


"""
partypes.py

This file contains special classes for TouchDesigner parameters.
These objects don't exist in TD but are necessary for tdi to 
provide proper return types from parameters.
"""

from typing import Any
from abc import abstractmethod

import typing as _T
ParValueT = _T.TypeVar('ParValueT')
###

class Par(_T.Generic[ParValueT]):
    """
    This is only a bandaid and nor for actual production use.
    """
    val:ParValueT

    @abstractmethod
    def eval(self) -> ParValueT:
        pass
	

class ParStr(Par["str"]):
	"TD Str Parameter"

class ParPython(Par["Any"]):
	"TD Python Parameter"

class ParFloat(Par["float"]):
	"TD Float Parameter"

class ParInt(Par["int"]):
	"TD Int Parameter"

class ParToggle(Par["bool"]):
	"TD Toggle Parameter"

class ParMomentary(Par["bool"]):
	"TD Momentary Parameter"

class ParPulse(Par["bool"]):
	"TD Pulse Parameter"

class ParMenu(Par["str"]):
	"TD Menu Parameter"

class ParStrMenu(Par["str"]):
	"TD StrMenu Parameter"

class ParRGB(Par["float"]):
	"TD RGB Parameter"

class ParRGBA(Par["float"]):
	"TD RGBA Parameter"

class ParUV(Par["float"]):
	"TD UV Parameter"

class ParUVW(Par["float"]):
	"TD UVW Parameter"

class ParWH(Par["float"]):
	"TD WH Parameter"

class ParXY(Par["float"]):
	"TD XY Parameter"

class ParXYZ(Par["float"]):
	"TD XYZ Parameter"

class ParXYZW(Par["float"]):
	"TD XYZW Parameter"

class ParObject(Par["None | ObjectCOMP"]):
	"TD Object Parameter"

class ParSOP(Par["None | SOP"]):
	"TD SOP Parameter"

class ParPOP(Par["None | POP"]):
	"TD POP Parameter"

class ParMAT(Par["None | MAT"]):
	"TD MAT Parameter"

class ParCHOP(Par["None | CHOP"]):
	"TD CHOP Parameter"

class ParTOP(Par["None | TOP"]):
	"TD TOP Parameter"

class ParDAT(Par["None | DAT"]):
	"TD DAT Parameter"

class ParPanelCOMP(Par["None | PanelCOMP"]):
	"TD PanelCOMP Parameter"

class ParCOMP(Par["None | COMP"]):
	"TD COMP Parameter"

class ParOP(Par["None | OP"]):
	"TD OP Parameter"

class ParFile(Par["str"]):
	"TD File Parameter"

class ParFileSave(Par["str"]):
	"TD FileSave Parameter"

class ParFolder(Par["str"]):
	"TD Folder Parameter"

class ParHeader(Par["str"]):
	"TD Header Parameter"

class ParSequence(Par["int"]):
	"TD Sequence Parameter"

class ParDATAdder(Par["None"]):
	"TD DATAdder Parameter"



from td import * # pyright: ignore[reportMissingImports]
from typing import cast
from pathlib import Path

def better_import():
    """
        Opiniated utility to import external data. It will create the correct operator for the filetype, set the refference to the file
        and lets you place it in the according location.
        If the path is relative, relative pathing will be used.
        If it is a tox, a hard refference to the tox will be created.
    """
    current_pane:Pane = ui.panes.current

    if not isinstance( current_pane, NetworkEditor ):
        ui.messageBox("Invalid Pane", "Requires NetworkEditor to be selected as current pane.")
        return
    
    current_editor:NetworkEditor = current_pane
    owner = current_pane.owner    
    _filepath = ui.chooseFile()
    if _filepath is None: return
    try:
        filepath = Path( _filepath ).absolute().relative_to( Path(".").absolute() )
    except ValueError:
        filepath = Path( _filepath )

    file_item = tdu.FileInfo( _filepath )
    if file_item.fileType in ("movie", "image"):
        new_op = owner.create( moviefileinTOP, "moviefilein1" )
        new_op.par.file.val = filepath
    elif file_item.fileType in ("audio",):
        new_op = owner.create( audiofileinCHOP, "audiofilein1" )
        new_op.par.file.val = filepath

    elif file_item.fileType in ("pointdata",):
        new_op = owner.create( pointfileinTOP, "pointdatain1" )
        new_op.par.file.val = filepath

    elif file_item.fileType in ("text",):
        new_op = owner.create( textDAT, "text1" )
        new_op.par.file.val = filepath
        new_op.par.language.menuIndex = 1
        new_op.par.extension.val = Path( filepath ).suffix.strip(".")
        new_op.par.loadonstartpulse.pulse()

    elif file_item.fileType in ("component",):
        new_op = cast(COMP, owner.loadTox( filepath ) )
        new_op.par.externaltox.val = filepath
        new_op.par.enableexternaltox.val = True
    else:
        return

    current_editor.placeOPs([new_op])
    


from datetime import datetime

class TimeTracker:
    def __init__(self, print_on_exit = False, print_method = print):
        self.print_on_exit = print_on_exit
        self.start = datetime.now()
        self.end = datetime.now()
        self.prin_method = print_method

    def __enter__(self):
        self.start = datetime.now()
    
    def __exit__(self, type, value, traceback):
        self.end = datetime.now()
        if self.print_on_exit: self.prin_method( f"Meassured Time in microseconds: {self.microseconds}" )

    @property
    def meassurement(self):
        return self.end - self.start
    
    @property
    def milliseconds(self):
        return self.meassurement.microseconds * 1000
    
    @property
    def microseconds(self):
        return self.meassurement.microseconds


from td import * # pyright: ignore[reportMissingImports]


from typing import Callable, List

Timer_Callback = Callable[["Timer"], None]

class Timer:
    def __init__(self, callbacks:list[Timer_Callback] = [], stepsize:int = 1, loop = True) -> None:
        self.stepsize = stepsize
        self.loop = loop
        self.callbacks = list() + callbacks
        self._runner:Run
        self.active = True
        self.start()
    
    def __del__(self):
        if not isinstance( self._runner, Run): return
        try:
            self._runner.kill()
        except tdError:
            pass

    callbacks:List[Timer_Callback] = list()

    def start(self):
        self._runner = run( "args[0]()", self.tick, delayFrames=self.stepsize )

    def tick(self):
        if self.active:
            for callback in self.callbacks:
                try:
                    callback(self)
                except Exception as e:
                    debug("Error during timer callback", callback, e)
        
        if self.loop: self.start()

class Ticker(Timer):
    def __init__(self, callbacks: List[Timer_Callback] = [], stepsize: int = 1) -> None:
        super().__init__(callbacks, stepsize, True)




import socket

def check_port(portnumber:int):
	with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as socket_probe:
		return bool( socket_probe.connect_ex( ( "127.0.0.1", int(portnumber))))

import socketserver
def get_free_port():
	with socketserver.TCPServer(("localhost", 0), socketserver.BaseRequestHandler ) as s:
		free_port = s.server_address[1]
	return free_port

import platform
def get_hostname():
	return platform.node()


import uuid
def get_computer_id():
	return str( uuid.getnode() )


from td import *  # pyright: ignore[reportMissingImports]

from typing import cast, TypeVar, Union, TypeGuard, Tuple, Type
T = TypeVar("T")
NoneType = type(None)

# Utils
class WrongOpType(TypeError):
	pass

def is_tuple(value) -> TypeGuard[Tuple]:
	return isinstance( value, tuple )

# Method Definitions.

def op_as(pattern:Union[str, int, Tuple[Union[str, int], ...]], asType:Type[T], includeUtility = False):
	"""
		Returns the operator as the defined asType to make use of strict typehinting.
		Idealy use the assert methods to make sure that this is actually true!
	"""
	if is_tuple( pattern ): _pattern = pattern
	else: _pattern = (pattern,)
	result = op(*_pattern, includeUtility = includeUtility) # pyright: ignore[reportArgumentType]
	return cast( T, result) 



def op_ex_as(pattern:Union[str, int, Tuple[Union[str, int], ...]], asType:Type[T], includeUtility = False): 
	"""
		Returns the operator as the defined asType to make use of strict typehinting.
		Raises an Exception when no operator is found.
		Idealy use the assert methods to make sure that this is actually true!
	"""
	if is_tuple( pattern ): _pattern = pattern
	else: _pattern = (pattern,)
	result = opex(*_pattern, includeUtility = includeUtility) # pyright: ignore[reportArgumentType]
	
	return cast( T, result) 

def op_assert(pattern:Union[str, int, Tuple[Union[str, int], ...]], asType:Type[T], includeUtility = False):
	"""
		Checks if the result is actually of the given type and raises WronfOpType if this is not the case. May return None.
	"""
	target_operator = op_as( pattern, asType, includeUtility=includeUtility)
	if not isinstance( target_operator, (asType, NoneType)):
		raise WrongOpType(f"Wrong operator type. Expected {asType}, got {type(target_operator)}")
	return target_operator

def op_ex_assert(pattern:Union[str, int, Tuple[Union[str, int], ...]], asType:Type[T], includeUtility = False):
	"""
		Uses the opEx method instead of op, so if the target op is None an exceptions raised from the call itself.
	"""
	target_operator = op_as( pattern, asType, includeUtility=includeUtility)
	if not isinstance( target_operator, asType):
		raise WrongOpType(f"Wrong operator type. Expected {asType}, got {type(target_operator)}")
	return target_operator


# This can stay for a moment.

opAs = op_as
opAsEx = op_ex_as



from td import *  # pyright: ignore[reportMissingImports]

from pathlib import Path
from typing import cast, TypeVar
from os import environ

T = TypeVar("T")


def refresh_tox(target_operator:COMP):
    target_operator.par.enableexternaltoxpulse.pulse(True)



def ensure_tox(filepath, op_shortcut, root_comp = root, default_path = "utils", reloadcustom = False, reloadbuiltin = False):
	if (_potentialy:= getattr(op, op_shortcut, None)) is not None:
		return _potentialy

	current_comp = root_comp
	for path_element in environ.get("ENSURE_UTILITY_PATH", default_path ).strip("/ ").split("/"):
		current_comp = current_comp.op( path_element ) or current_comp.create( baseCOMP, path_element)

	newly_loaded 							= current_comp.loadTox(filepath)
	newly_loaded.name 						= op_shortcut
	newly_loaded.par.opshortcut.val 		= op_shortcut
	newly_loaded.par.externaltox.val 		= Path(filepath).relative_to( Path(".").absolute() )
	newly_loaded.par.enableexternaltox.val 	= True
	newly_loaded.par.savebackup.val 		= False
	newly_loaded.par.reloadcustom.val 		= reloadcustom
	newly_loaded.par.reloadbuiltin.val 		= reloadbuiltin
	newly_loaded.par.enableexternaltoxpulse.pulse()
	
	return newly_loaded

def iter_parents( target_op:OP ):
    next_parent = target_op
    while next_parent:
        next_parent = target_op.parent()
        if next_parent is None: break
        target_op = next_parent
        yield target_op
    return target_op
